@@ -1,7 +1,7 @@
 //! Contains structs and traits closely related to D-Bus messages.
 
 use std::{fmt, ptr};
-use super::{ffi, Error, libc, to_c_str, c_str_to_slice, init_dbus};
+use super::{ffi, Error, libc, init_dbus};
 use crate::strings::{BusName, Path, Interface, Member, ErrorName};
 use std::ffi::CStr;
 
@@ -42,7 +42,7 @@ impl Message {
         init_dbus();
         let (d, p, i, m) = (destination.into(), path.into(), iface.into(), method.into());
         let ptr = unsafe {
-            ffi::dbus_message_new_method_call(d.as_ref().as_ptr(), p.as_ref().as_ptr(), i.as_ref().as_ptr(), m.as_ref().as_ptr())
+            ffi::dbus_message_new_method_call(d.as_ptr(), p.as_ptr(), i.as_ptr(), m.as_ptr())
         };
         if ptr.is_null() { Err("D-Bus error: dbus_message_new_method_call failed".into()) }
         else { Ok(Message { msg: ptr}) }
@@ -52,13 +52,27 @@ impl Message {
     pub fn method_call(destination: &BusName, path: &Path, iface: &Interface, name: &Member) -> Message {
         init_dbus();
         let ptr = unsafe {
-            ffi::dbus_message_new_method_call(destination.as_ref().as_ptr(), path.as_ref().as_ptr(),
-                iface.as_ref().as_ptr(), name.as_ref().as_ptr())
+            ffi::dbus_message_new_method_call(destination.as_ptr(), path.as_ptr(),
+                iface.as_ptr(), name.as_ptr())
         };
         if ptr.is_null() { panic!("D-Bus error: dbus_message_new_method_call failed") }
         Message { msg: ptr}
     }
 
+    /// Creates a new message that is a replica of this message, but without a serial.
+    ///
+    /// May fail if out of memory or file descriptors.
+    pub fn duplicate(&self) -> Result<Self, String> {
+        let ptr = unsafe {
+            ffi::dbus_message_copy(self.msg)
+        };
+        if ptr.is_null() {
+            Err("D-Bus error: dbus_message_copy failed".into())
+        } else {
+            Ok(Message { msg: ptr })
+        }
+    }
+
     /// Creates a new method call message.
     pub fn call_with_args<'d, 'p, 'i, 'm, A, D, P, I, M>(destination: D, path: P, iface: I, method: M, args: A) -> Message
     where D: Into<BusName<'d>>, P: Into<Path<'p>>, I: Into<Interface<'i>>, M: Into<Member<'m>>, A: AppendAll {
@@ -69,7 +83,7 @@ impl Message {
 
     /// Creates a new signal message.
     pub fn new_signal<P, I, M>(path: P, iface: I, name: M) -> Result<Message, String>
-    where P: Into<Vec<u8>>, I: Into<Vec<u8>>, M: Into<Vec<u8>> {
+    where P: Into<String>, I: Into<String>, M: Into<String> {
         init_dbus();
 
         let p = Path::new(path)?;
@@ -77,7 +91,7 @@ impl Message {
         let m = Member::new(name)?;
 
         let ptr = unsafe {
-            ffi::dbus_message_new_signal(p.as_ref().as_ptr(), i.as_ref().as_ptr(), m.as_ref().as_ptr())
+            ffi::dbus_message_new_signal(p.as_ptr(), i.as_ptr(), m.as_ptr())
         };
         if ptr.is_null() { Err("D-Bus error: dbus_message_new_signal failed".into()) }
         else { Ok(Message { msg: ptr}) }
@@ -87,7 +101,7 @@ impl Message {
     pub fn signal(path: &Path, iface: &Interface, name: &Member) -> Message {
         init_dbus();
         let ptr = unsafe {
-            ffi::dbus_message_new_signal(path.as_ref().as_ptr(), iface.as_ref().as_ptr(), name.as_ref().as_ptr())
+            ffi::dbus_message_new_signal(path.as_ptr(), iface.as_ptr(), name.as_ptr())
         };
         if ptr.is_null() { panic!("D-Bus error: dbus_message_new_signal failed") }
         Message { msg: ptr}
@@ -115,17 +129,9 @@ impl Message {
         m
     }
 
-    /// The old way to create a new error reply
-    #[deprecated]
-    pub fn new_error(m: &Message, error_name: &str, error_message: &str) -> Option<Message> {
-        let (en, em) = (to_c_str(error_name), to_c_str(error_message));
-        let ptr = unsafe { ffi::dbus_message_new_error(m.msg, en.as_ptr(), em.as_ptr()) };
-        if ptr.is_null() { None } else { Some(Message { msg: ptr} ) }
-    }
-
     /// Creates a new error reply
     pub fn error(&self, error_name: &ErrorName, error_message: &CStr) -> Message {
-        let ptr = unsafe { ffi::dbus_message_new_error(self.msg, error_name.as_ref().as_ptr(), error_message.as_ptr()) };
+        let ptr = unsafe { ffi::dbus_message_new_error(self.msg, error_name.as_ptr(), error_message.as_ptr()) };
         if ptr.is_null() { panic!("D-Bus error: dbus_message_new_error failed") }
         Message { msg: ptr}
     }
@@ -352,9 +358,10 @@ impl Message {
         }
     }
 
-    fn msg_internal_str<'a>(&'a self, c: *const libc::c_char) -> Option<&'a [u8]> {
-        if c.is_null() { None }
-        else { Some( unsafe { CStr::from_ptr(c) }.to_bytes_with_nul()) }
+    fn msg_internal_str<'a>(&'a self, c: *const libc::c_char) -> Option<&'a str> {
+        if c.is_null() { return None };
+        let cc = unsafe { CStr::from_ptr(c) };
+        std::str::from_utf8(cc.to_bytes_with_nul()).ok()
     }
 
     /// Gets the name of the connection that originated this message.
@@ -363,18 +370,6 @@ impl Message {
             .map(|s| unsafe { BusName::from_slice_unchecked(s) })
     }
 
-    /// Returns a tuple of (Message type, Path, Interface, Member) of the current message.
-    #[deprecated]
-    pub fn headers(&self) -> (MessageType, Option<String>, Option<String>, Option<String>) {
-        let p = unsafe { ffi::dbus_message_get_path(self.msg) };
-        let i = unsafe { ffi::dbus_message_get_interface(self.msg) };
-        let m = unsafe { ffi::dbus_message_get_member(self.msg) };
-        (self.msg_type(),
-         c_str_to_slice(&p).map(|s| s.to_string()),
-         c_str_to_slice(&i).map(|s| s.to_string()),
-         c_str_to_slice(&m).map(|s| s.to_string()))
-    }
-
     /// Gets the object path this Message is being sent to.
     pub fn path(&self) -> Option<Path> {
         self.msg_internal_str(unsafe { ffi::dbus_message_get_path(self.msg) })
@@ -430,13 +425,53 @@ impl Message {
         Message { msg: ptr }
     }
 
-    /// Sets serial number manually - should not be used in production code
+    /// Sets serial number manually - mostly for internal use
     ///
-    /// However, this can be very useful in test code that is supposed to handle a method call.
+    /// When sending a message, a serial will be automatically assigned, so you don't need to call
+    /// this method. However, it can be very useful in test code that is supposed to handle a method call.
     /// This way, you can create a method call and handle it without sending it to a real D-Bus instance.
     pub fn set_serial(&mut self, val: u32) {
         unsafe { ffi::dbus_message_set_serial(self.msg, val) };
     }
+
+    /// Marshals a message - mostly for internal use
+    ///
+    /// The function f will be called one or more times with bytes to be written somewhere.
+    /// You should call set_serial to manually set a serial number before calling this function
+    pub fn marshal<E, F: FnMut(&[u8]) -> Result<(), E>>(&self, mut f: F) -> Result<(), E> {
+        let mut len = 0;
+        let mut data = ptr::null_mut();
+        if unsafe { ffi::dbus_message_marshal(self.msg, &mut data, &mut len) } == 0 {
+            panic!("out of memory");
+        }
+        let s = unsafe { std::slice::from_raw_parts(data as *mut u8 as *const u8, len as usize) };
+        let r = f(s);
+        unsafe { ffi::dbus_free(data as *mut _) };
+        r
+    }
+
+    /// Demarshals a message - mostly for internal use
+    pub fn demarshal(data: &[u8]) -> Result<Self, Error> {
+        let mut e = Error::empty();
+        let p = unsafe { ffi::dbus_message_demarshal(data.as_ptr() as *const _, data.len() as _, e.get_mut()) };
+        if p == ptr::null_mut() {
+            Err(e)
+        } else {
+            Ok(Self::from_ptr(p, false))
+        }
+    }
+
+    /// Returns the size of the message - mostly for internal use
+    ///
+    /// Returns Err(()) on protocol errors. Make sure you have at least 16 bytes in the buffer
+    /// before calling this method.
+    pub fn demarshal_bytes_needed(data: &[u8]) -> Result<usize, ()> {
+        const MIN_HEADER: usize = 16;
+        if data.len() < MIN_HEADER { return Ok(MIN_HEADER); }
+        let x = unsafe { ffi::dbus_message_demarshal_bytes_needed(data.as_ptr() as *const _, data.len() as _) };
+        if x < MIN_HEADER as _ { Err(()) } else { Ok(x as usize) }
+    }
+
 }
 
 impl Drop for Message {
@@ -486,4 +521,16 @@ mod test {
         m.set_no_reply(true);
         assert!(m.get_no_reply());
     }
+
+    #[test]
+    fn marshal() {
+        let mut m = Message::new_method_call("org.freedesktop.DBus", "/org/freedesktop/DBus", "org.freedesktop.DBus", "Hello").unwrap();
+        m.set_serial(1);
+        let r = m.marshal(|d| {
+            let m2 = Message::demarshal(d).unwrap();
+            assert_eq!(&*m2.path().unwrap(), "/org/freedesktop/DBus");
+            Err(45)
+        });
+        assert_eq!(45, r.unwrap_err());
+    }
 }
