@@ -17,7 +17,7 @@ use crate::{Error, Message};
 use crate::channel::{MatchingReceiver, Channel, Sender, Token};
 use crate::strings::{BusName, Path, Interface, Member};
 use crate::arg::{AppendAll, ReadAll, IterAppend};
-use crate::message::MatchRule;
+use crate::message::{MatchRule, MessageType};
 
 use std::sync::{Arc, Mutex};
 use std::sync::atomic::{AtomicUsize, Ordering::SeqCst};
@@ -75,6 +75,7 @@ pub struct LocalConnection {
     replies: RefCell<Replies<LocalRepliesCb>>,
     timeout_maker: Option<TimeoutMakerCb>,
     waker: Option<WakerCb>,
+    all_signal_matches: bool,
 }
 
 /// A connection to D-Bus, async version, which is Send but not Sync.
@@ -84,6 +85,7 @@ pub struct Connection {
     replies: RefCell<Replies<RepliesCb>>,
     timeout_maker: Option<TimeoutMakerCb>,
     waker: Option<WakerCb>,
+    all_signal_matches: bool,
 }
 
 /// A connection to D-Bus, Send + Sync + async version
@@ -93,6 +95,7 @@ pub struct SyncConnection {
     replies: Mutex<Replies<SyncRepliesCb>>,
     timeout_maker: Option<TimeoutMakerCb>,
     waker: Option<WakerCb>,
+    all_signal_matches: bool,
 }
 
 use stdintf::org_freedesktop_dbus::DBus;
@@ -113,6 +116,7 @@ impl From<Channel> for $c {
             filters: Default::default(),
             timeout_maker: None,
             waker: None,
+            all_signal_matches: false,
         }
     }
 }
@@ -122,7 +126,17 @@ impl AsRef<Channel> for $c {
 }
 
 impl Sender for $c {
-    fn send(&self, msg: Message) -> Result<u32, ()> { self.channel.send(msg) }
+    fn send(&self, msg: Message) -> Result<u32, ()> {
+        let token = self.channel.send(msg);
+        if self.channel.has_messages_to_send() {
+            // Non-blocking send failed
+            // Wake up task that will send the message
+            if self.waker.as_ref().map(|wake| wake().is_err() ).unwrap_or(false) {
+                return Err(());
+            }
+        }
+        token
+    }
 }
 
 impl MatchingReceiver for $c {
@@ -178,13 +192,29 @@ impl Process for $c {
                 return;
             }
         }
-        let ff = self.filters_mut().remove_matching(&msg);
-        if let Some(mut ff) = ff {
-            if ff.2(msg, self) {
-                self.filters_mut().insert(ff);
+        if self.all_signal_matches && msg.msg_type() == MessageType::Signal {
+            // If it's a signal and the mode is enabled, send a copy of the message to all
+            // matching filters.
+            for mut ff in self.filters_mut().remove_all_matching(&msg) {
+                if let Ok(copy) = msg.duplicate() {
+                    if ff.2(copy, self) {
+                        self.filters_mut().insert(ff);
+                    }
+                } else {
+                    // Silently drop the message, but add the filter back.
+                    self.filters_mut().insert(ff);
+                }
+            }
+        } else {
+            // Otherwise, send the original message to only the first matching filter.
+            let ff = self.filters_mut().remove_first_matching(&msg);
+            if let Some(mut ff) = ff {
+                if ff.2(msg, self) {
+                    self.filters_mut().insert(ff);
+                }
+            } else if let Some(reply) = crate::channel::default_reply(&msg) {
+                let _ = self.channel.send(reply);
             }
-        } else if let Some(reply) = crate::channel::default_reply(&msg) {
-            let _ = self.send(reply);
         }
     }
 }
@@ -228,6 +258,10 @@ impl $c {
 
     /// Adds a new match to the connection, and sets up a callback when this message arrives.
     ///
+    /// If multiple [`MatchRule`]s match the same message, then by default only the first will get
+    /// the callback. This behaviour can be changed for signal messages by calling
+    /// [`set_signal_match_mode`](Self::set_signal_match_mode).
+    ///
     /// The returned value can be used to remove the match.
     pub async fn add_match(&self, match_rule: MatchRule<'static>) -> Result<MsgMatch, Error> {
         let m = match_rule.match_str();
@@ -260,6 +294,21 @@ impl $c {
         let (mr, _) = self.stop_receive(id).ok_or_else(|| Error::new_failed("No match with that id found"))?;
         self.remove_match_no_cb(&mr.match_str()).await
     }
+
+    /// If true, configures the connection to send signal messages to all matching [`MatchRule`]
+    /// filters added with [`add_match`](Self::add_match) rather than just the first one. This comes
+    /// with the following gotchas:
+    ///
+    ///  * The messages might be duplicated, so the message serial might be lost (this is
+    ///    generally not a problem for signals).
+    ///  * Panicking inside a match callback might mess with other callbacks, causing them
+    ///    to be permanently dropped.
+    ///  * Removing other matches from inside a match callback is not supported.
+    ///
+    /// This is false by default, for a newly-created connection.
+    pub fn set_signal_match_mode(&mut self, match_all: bool) {
+        self.all_signal_matches = match_all;
+    }
 }
 
 
@@ -383,8 +432,8 @@ impl MsgMatch {
     /// Note: If the receiving end is disconnected and a message is received,
     /// the message matching will end but not in a clean fashion. Call remove_match() to
     /// stop matching cleanly.
-    pub fn msg_stream(self) -> (Self, futures::channel::mpsc::UnboundedReceiver<Message>) {
-        let (sender, receiver) = futures::channel::mpsc::unbounded();
+    pub fn msg_stream(self) -> (Self, futures_channel::mpsc::UnboundedReceiver<Message>) {
+        let (sender, receiver) = futures_channel::mpsc::unbounded();
         (self.msg_cb(move |msg| {
             sender.unbounded_send(msg).is_ok()
         }), receiver)
@@ -406,8 +455,8 @@ impl MsgMatch {
     ///    async {}
     /// });
     /// ```
-    pub fn stream<R: ReadAll + Send + 'static>(self) -> (Self, futures::channel::mpsc::UnboundedReceiver<(Message, R)>) {
-        let (sender, receiver) = futures::channel::mpsc::unbounded();
+    pub fn stream<R: ReadAll + Send + 'static>(self) -> (Self, futures_channel::mpsc::UnboundedReceiver<(Message, R)>) {
+        let (sender, receiver) = futures_channel::mpsc::unbounded();
         (self.cb(move |msg, r| {
             sender.unbounded_send((msg, r)).is_ok()
         }), receiver)
@@ -449,7 +498,7 @@ struct MRAwait {
 }
 
 async fn method_call_await(mra: MRAwait) -> Result<Message, Error> {
-    use futures::future;
+    use futures_util::future;
     let MRAwait { mrouter, token, timeout, timeoutfn } = mra;
     if token.is_err() { return Err(Error::new_failed("Failed to send message")) };
     let timeout = if let Some(tfn) = timeoutfn { tfn(timeout) } else { Box::pin(future::pending()) };
@@ -487,7 +536,7 @@ where
         args.append(&mut IterAppend::new(&mut msg));
         let mra = self.method_call_setup(msg);
         let r = method_call_await(mra);
-        let r = futures::FutureExt::map(r, |r| -> Result<R, Error> { r.and_then(|rmsg| rmsg.read_all()) } );
+        let r = futures_util::FutureExt::map(r, |r| -> Result<R, Error> { r.and_then(|rmsg| rmsg.read_all()) } );
         MethodReply::new(r)
     }
 }
@@ -507,7 +556,7 @@ impl Future for MROuter {
         let r = mem::replace(&mut *inner, MRInner::Neither);
         if let MRInner::Ready(r) = r { task::Poll::Ready(r) }
         else {
-            mem::replace(&mut *inner, MRInner::Pending(ctx.waker().clone()));
+            *inner = MRInner::Pending(ctx.waker().clone());
             return task::Poll::Pending
         }
     }
